/*
 * BrainFlak (https://esolangs.org/wiki/Brain-Flak) interpreter in Rust.
 * Values are signed 64-bit integers or larger.
 * No syntax errors. Any valid opening bracket can be closed with any valid
 * closing bracket, but any combination not specified below causes undefined
 * behavior when evaluated. To improve user experience, this implementation
 * treats those the same way as {]. Any unclosed brackets, along with their
 * contents, or extraneous closing brackets are discarded.
 *
 * Custom operations:
 *
 *    {], {...] - Evaluates to 0. The inner expression, if any, is not
 *    evaluated. Can be used for comments, as long as opening brackets inside
 *    are balanced with an equal amount of closing brackets.
 *
 *    <) - Read a byte from standard input. Evaluates to -1 on EOF, -2 on error,
 *    and a positive integer (the byte) if successful.
 *
 *    <...) - If the inner expression evaluates to a non-negative value,
 *    truncate it to the size of a byte and write it to the standard output.
 *    Negative values cause undefined behavior, except -1 and -2 which close the
 *    standard output and input, respectively (does not actually work properly
 *    in this implementation as Rust doesn't have built-in constructs to do
 *    this). Evaluates to 0 on success, 1 on failure.
 */

/*
 * Copyright (C)2023 Anonymous
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

mod brainflak {
    #[derive(Copy, Clone, PartialEq, Eq)]
    enum BracketType {
        Round,
        Square,
        Curly,
        Angle,
    }
    #[derive(Copy, Clone)]
    enum BracketSide {
        Opening,
        Closing,
    }
    #[derive(Copy, Clone)]
    struct Bracket {
        btype: BracketType,
        side: BracketSide,
    }
    use BracketSide::*;
    use BracketType::*;
    struct Func {
        btype: BracketType,
        monad: bool,
        addr: usize,
    }
    impl TryFrom<u8> for Bracket {
        type Error = ();
        fn try_from(c: u8) -> Result<Self, ()> {
            match c {
                b'(' => Ok(Self {
                    btype: Round,
                    side: Opening,
                }),
                b'[' => Ok(Self {
                    btype: Square,
                    side: Opening,
                }),
                b'{' => Ok(Self {
                    btype: Curly,
                    side: Opening,
                }),
                b'<' => Ok(Self {
                    btype: Angle,
                    side: Opening,
                }),
                b')' => Ok(Self {
                    btype: Round,
                    side: Closing,
                }),
                b']' => Ok(Self {
                    btype: Square,
                    side: Closing,
                }),
                b'}' => Ok(Self {
                    btype: Curly,
                    side: Closing,
                }),
                b'>' => Ok(Self {
                    btype: Angle,
                    side: Closing,
                }),
                _ => Err(()),
            }
        }
    }
    pub struct Program {
        code: Box<[u8]>,
    }
    mod ops {
        pub const NOP: u8 = 0;

        pub const ONE: u8 = 1;
        pub const HEIGHT: u8 = 2;
        pub const POP: u8 = 3;
        pub const SWAP: u8 = 4;

        pub const NEW: u8 = 5;
        pub const PUSH: u8 = 6;
        pub const NEGATE: u8 = 7;
        pub const DROP: u8 = 8;

        pub const SAVE: u8 = 9;
        pub const ZRET: u8 = 10;

        pub const READ: u8 = 11;
        pub const WRITE: u8 = 12;
    }
    use ops::*;
    use std::io;
    use std::io::ErrorKind;
    use std::io::{Read, Write};
    use std::mem;
    impl Program {
        fn backtrack(code: &mut Vec<u8>, f: &Func) {
            code.truncate(f.addr);
        }
        pub fn run(&self, input: Vec<i64>, stdin: impl Read, stdout: impl Write) -> Vec<i64> {
            let mut active = input;
            let mut passive: Vec<i64> = Vec::new();
            let mut data: Vec<i64> = Vec::new();
            let mut rets: Vec<u32> = Vec::new();
            let mut exec: u32 = 0;
            let mut stdin = Some(stdin);
            let mut stdout = Some(stdout);
            let clen = self.code.len() as u32;
            while exec < clen {
                let c = self.code[exec as usize];
                match c {
                    SAVE | ZRET => match c {
                        SAVE => {
                            rets.push(exec + 5);
                            data.push(0);
                            exec = u32::from_le_bytes(
                                self.code[((exec + 1) as usize)..((exec + 5) as usize)]
                                    .try_into()
                                    .unwrap(),
                            );
                        }
                        ZRET => {
                            if active.last().map(|v| *v).unwrap_or(0) != 0 {
                                exec = *rets.last().unwrap();
                            } else {
                                let v = data.pop().unwrap();
                                data.last_mut().map(|pc| *pc += v);
                                rets.pop().unwrap();
                                exec += 1;
                            };
                        }
                        _ => unreachable!(),
                    },
                    NEW => {
                        data.push(0);
                        exec += 1;
                    }
                    NOP | ONE | HEIGHT | POP | SWAP | READ => {
                        let v = match c {
                            NOP => 0,
                            ONE => 1,
                            HEIGHT => active.len() as i64,
                            POP => active.pop().unwrap_or(0),
                            SWAP => {
                                mem::swap(&mut active, &mut passive);
                                0
                            }
                            READ => {
                                let mut buffer = [0; 1];
                                match stdin.as_mut() {
                                    Some(stdin) => match stdin.read_exact(&mut buffer) {
                                        Ok(_) => buffer[0] as i64,
                                        Err(e) => match e.kind() {
                                            ErrorKind::UnexpectedEof => -1,
                                            _ => -2,
                                        },
                                    },
                                    None => -2,
                                }
                            }
                            _ => unreachable!(),
                        };
                        data.last_mut().map(|pc| *pc += v);
                        exec += 1;
                    }
                    PUSH | NEGATE | DROP | WRITE => {
                        let mut v = data.pop().unwrap();
                        match c {
                            PUSH => active.push(v),
                            NEGATE => v = -v,
                            DROP => v = 0,
                            WRITE => {
                                v = match v {
                                    v if v >= 0 => match stdout.as_mut() {
                                        Some(stdout) => match stdout.write_all(&[v as u8]) {
                                            Ok(_) => 0,
                                            Err(_) => 1,
                                        },
                                        None => 1,
                                    },
                                    -1 => stdin.take().map_or(1, |_| 0),
                                    -2 => stdout.take().map_or(1, |_| 0),
                                    _ => 1,
                                }
                            }
                            _ => unreachable!(),
                        }
                        data.last_mut().map(|pv| *pv += v);
                        exec += 1;
                    }
                    _ => panic!(),
                }
            }
            active
        }
        pub fn new(file: impl Read) -> io::Result<Self> {
            let mut funcs: Vec<Func> = Vec::new();
            let mut code: Vec<u8> = Vec::new();
            for b in file.bytes().filter_map(|c| match c {
                Ok(c) => Bracket::try_from(c).ok().map(|c| Ok(c)),
                Err(e) => Some(Err(e)),
            }) {
                let b = b?;
                match b.side {
                    Opening => {
                        funcs.last_mut().map(|f| {
                            if f.monad {
                                return;
                            }
                            f.monad = true;
                            if f.btype == Curly {
                                code.extend([0; 4]);
                            }
                        });
                        funcs.push(Func {
                            btype: b.btype,
                            monad: false,
                            addr: code.len(),
                        });
                        code.push(0);
                    }
                    Closing => 'm: {
                        let f = if let Some(f) = funcs.pop() {
                            f
                        } else {
                            break 'm;
                        };
                        'e: {
                            match f.monad {
                                false => {
                                    code[f.addr] = match (f.btype, b.btype) {
                                        (Round, Round) => ONE,
                                        (Square, Square) => HEIGHT,
                                        (Curly, Curly) => POP,
                                        (Angle, Angle) => SWAP,
                                        (Angle, Round) => READ,
                                        _ => break 'e Err(()),
                                    }
                                }
                                true => {
                                    if let (Curly, Curly) = (f.btype, b.btype) {
                                        code[f.addr] = SAVE;
                                        let addr = &(code.len() as u32).to_le_bytes();
                                        code[(f.addr + 1)..(f.addr + 5)].copy_from_slice(addr);
                                        code.push(ZRET);
                                    } else {
                                        code[f.addr] = NEW;
                                        code.push(match (f.btype, b.btype) {
                                            (Round, Round) => PUSH,
                                            (Square, Square) => NEGATE,
                                            (Angle, Angle) => DROP,
                                            (Angle, Round) => WRITE,
                                            _ => break 'e Err(()),
                                        });
                                    }
                                }
                            }
                            Ok(())
                        }
                        .err()
                        .map(|_| Self::backtrack(&mut code, &f));
                    }
                }
                if code.len() > (u32::MAX as usize) {
                    panic!("Program too large");
                }
            }
            funcs.first().map(|f| Self::backtrack(&mut code, &f));
            code.shrink_to_fit();
            Ok(Self {
                code: code.into_boxed_slice(),
            })
        }
    }
}

pub use brainflak::*;
use std::env;
use std::fs::File;
use std::io;
use std::str::FromStr;

fn main() {
    let mut args = env::args();
    args.next();
    let pfile = File::open(args.next().expect("invalid argument")).expect("failed to open file");
    let mut inputs = Vec::new();
    for p in args {
        inputs.push(i64::from_str(&p).expect("invalid argument"));
    }
    let program = Program::new(pfile).expect("fail");
    for i in program.run(inputs, io::stdin(), io::stdout()) {
        println!("{}", i);
    }
}
